import logging

from airflow.models.dag import DAG
from airflow.providers.google.cloud.transfers.gcs_to_bigquery import GCSToBigQueryOperator

from gcp_airflow_foundations.base_class.data_source_table_config import DataSourceTablesConfig

from gcp_airflow_foundations.source_class.source import DagBuilder

from gcp_airflow_foundations.operators.api.operators.twilio_operator import TwilioToBigQueryOperator
from gcp_airflow_foundations.operators.api.schemas.twilio import get_twilio_schema

from gcp_airflow_foundations.common.gcp.load_builder import load_builder

from urllib.parse import urlparse


class TwilioToBQDagBuilder(DagBuilder):
    """
    Builds DAGs to load data from Twilio's API to a BigQuery Table.
    """
    source_type = "TWILIO"

    def build_dags(self):
        data_source = self.config.source
        logging.info(f"Building DAG for Twilio {data_source.name}")
        project_id = data_source.gcp_project

        logging.info(f"Building DAG for Twilio {data_source.name}")

        # bq args
        landing_dataset = data_source.landing_zone_options.landing_zone_dataset

        dags = []
        for table_config in self.config.tables:
            table_default_task_args = self.default_task_args_for_table(
                self.config, table_config
            )
            logging.info(f"table_default_task_args {table_default_task_args}")

            start_date = table_default_task_args["start_date"]

            with DAG(
                dag_id=f"twilio_to_bq_{table_config.table_name}",
                description=f"BigQuery load for {table_config.table_name}",
                schedule_interval=None,
                default_args=table_default_task_args
            ) as dag:

                destination_table = f"{table_config.landing_zone_table_name_override}"

                # 1 Load Twilio data to BQ Landing Zone
                load_to_bq_landing = TwilioToBigQueryOperator(
                    task_id='import_twilio_to_bq_landing',
                    twilio_account_sid=data_source.extra_options["twilio_account_sid"],
                    project_id=project_id,
                    dataset_id=landing_dataset,
                    schema_fields=get_twilio_schema(tableId=destination_table),
                    table_id=destination_table,
                    dag=dag)

                #2 Create task groups for loading data to ODS/HDS tables
                taskgroups = load_builder(
                    project_id=data_source.gcp_project,
                    table_id=table_config.table_name,
                    dataset_id=data_source.dataset_data_name,
                    landing_zone_dataset=landing_dataset,
                    landing_zone_table_name_override=table_config.landing_zone_table_name_override,
                    surrogate_keys=table_config.surrogate_keys,
                    column_mapping=table_config.column_mapping,
                    gcs_schema_object=None,
                    schema_fields=get_twilio_schema(tableId=destination_table),
                    ods_table_config=table_config.ods_config,
                    hds_table_config=table_config.hds_config,
                    preceding_task=load_to_bq_landing,
                    dag=dag
                )

                logging.info(f"Created dag for {table_config}, {dag}")

                dags.append(dag)

        return dags

    def validate_extra_options(self):
        # Example of extra validations to do
        extra_options = self.config.source.extra_options

        # assert bucket and object/s are non-empty
        assert extra_options["gcs_bucket"]
        assert extra_options["twilio_account_sid"]


from airflow.exceptions import AirflowException

from dacite import Config
from dataclasses import dataclass, field

from datetime import datetime
from typing import List, Optional

from gcp_airflow_foundations.base_class.ods_metadata_config import OdsTableMetadataConfig

@dataclass
class OdsTableConfig:
    """
    Attributes:
        merge_type : Only SG_KEY_WITH_HASH is currently supported
        ods_metadata : See OdsTableMetadataConfig class 
    """
    ods_metadata: OdsTableMetadataConfig
    merge_type: str = "SG_KEY_WITH_HASH"



from gcp_airflow_foundations.enums.ingestion_type import IngestionType

class SqlHelperHDS:
    """
    SQL helper class used to formulate the SQL queries for the merge operations of HDS tables.
    
    :param source: Source table name
    :type source: str
    :param target: Target table name
    :type target: str
    :param source_dataset: Source dataset name
    :type source_dataset: str
    :param target_dataset: Target dataset name
    :type target_dataset: str
    :param surrogate_keys: Surrogate keys of the target table
    :type surrogate_keys: list
    :param columns: List of source table columns
    :type columns: list
    :param gcp_conn_id: Airflow GCP connection ID
    :type gcp_conn_id: str
    :param column_mapping: Column mapping
    :type column_mapping: dict
    :param time_partitioning: Time partitioning option for BigQuery target table. One of HOUR, DAY, or MONTH
    :type time_partitioning: str
    :param hds_metadata: User-provided options for HDS metadata column naming
    :type hds_metadata: HdsTableMetadataConfig
    """

    def __init__(
        self,
        source_dataset,
        target_dataset,
        source,
        target,
        columns,
        surrogate_keys,
        column_mapping,
        hds_metadata,
        time_partitioning=None,
        gcp_conn_id='google_cloud_default'):

        self.source_dataset = source_dataset
        self.target_dataset = target_dataset
        self.source = source
        self.target = target
        self.surrogate_keys = surrogate_keys
        self.column_mapping = column_mapping
        self.hds_metadata = hds_metadata
        self.gcp_conn_id = gcp_conn_id
        self.time_partitioning = time_partitioning
        self.columns = columns
        self.hash_column_name = hds_metadata.hash_column_name
        self.eff_start_time_column_name = hds_metadata.eff_start_time_column_name
        self.partition_column_name = hds_metadata.partition_time_column_name
        self.eff_end_time_column_name = hds_metadata.eff_end_time_column_name

        self.columns_str_source: str = ",".join(["`{}`".format(col) for col in columns])
        self.columns_str_keys: str = ",".join(surrogate_keys)
        self.columns_str_target: str = ",".join(["`{}`".format(column_mapping[i]) for i in columns])

    def create_scd2_sql_with_hash(self, ingestion_type):
        comma = ","

        TEMPLATE = """
                    MERGE `{target}` T
                    USING ( {source_query} ) S
                    ON {merge_condition}
                    WHEN MATCHED AND {search_condition} THEN {matched_clause}
                    WHEN NOT MATCHED BY TARGET THEN {not_matched_clause}
        """

        target = f"{self.target_dataset}.{self.target}"
        source_query = f"""
                SELECT  {",".join(["{} AS join_key_{}".format(surrogate_key, surrogate_key) for surrogate_key in self.surrogate_keys])},
                        {",".join([col for col in self.columns])}
                FROM `{self.source_dataset}.{self.source}`
                UNION ALL 
                SELECT
                    {",".join(["source.`{}`".format(surrogate_key) for surrogate_key in self.surrogate_keys])},
                    {",".join(["source.`{}`".format(col) if col not in self.surrogate_keys else "NULL" for col in self.columns])}
                    FROM `{self.source_dataset}.{self.source}` source
                    JOIN `{self.target_dataset}.{self.target}` target
                    ON {' AND '.join([f'target.{self.column_mapping[surrogate_key]}=source.{surrogate_key}' for surrogate_key in self.surrogate_keys])}
                    WHERE ( 
                            {' AND '.join([f'MD5(TO_JSON_STRING(target.`{self.column_mapping[col]}`)) != MD5(TO_JSON_STRING(source.`{col}`))' for col in self.columns])}
                            AND target.{self.eff_end_time_column_name} IS NULL
                        )
        """
        merge_condition = f"{' AND '.join([f'T.{self.column_mapping[surrogate_key]}=S.{surrogate_key}' for surrogate_key in self.surrogate_keys])}"
        search_condition = f"{' AND '.join([f'MD5(TO_JSON_STRING(T.`{self.column_mapping[col]}`)) != MD5(TO_JSON_STRING(S.`{col}`))' for col in self.columns])}"
        matched_clause = f"UPDATE SET {self.eff_end_time_column_name} = CURRENT_TIMESTAMP()"
        not_matched_clause = f"""
            INSERT ({self.columns_str_target}, {self.eff_start_time_column_name}, {self.eff_end_time_column_name}, {self.hash_column_name})
            VALUES ({",".join(["`{}`".format(col) if col not in self.surrogate_keys else "join_key_{}".format(col) for col in self.columns])}, CURRENT_TIMESTAMP(), NULL, TO_BASE64(MD5(TO_JSON_STRING(S))))
        """

        sql = TEMPLATE.format(
            target=target,
            source_query=source_query,
            merge_condition=merge_condition,
            search_condition=search_condition,
            matched_clause=matched_clause,
            not_matched_clause=not_matched_clause
        )

        if ingestion_type == IngestionType.INCREMENTAL:
            return sql

        elif ingestion_type == IngestionType.FULL:
            return sql + \
                f"""WHEN NOT MATCHED BY SOURCE THEN UPDATE
                        SET T.{self.eff_end_time_column_name} = CURRENT_TIMESTAMP()
                """  

    def create_snapshot_sql_with_hash(self, partition_timestamp):
        sql = f"""
                    SELECT
                        {(','.join(f'`{col}` AS `{self.column_mapping[col]}`' for col in self.columns))},
                        CURRENT_TIMESTAMP() AS {self.eff_start_time_column_name},
                        TIMESTAMP_TRUNC('{partition_timestamp}', {self.time_partitioning}) AS {self.partition_column_name},
                        TO_BASE64(MD5(TO_JSON_STRING(S))) AS {self.hash_column_name}
                    FROM {self.source_dataset}.{self.source} S
        """
        return sql


import datetime
import os
from typing import Any, Callable, FrozenSet, Iterable, Optional, Union
import re

from sqlalchemy import func

from airflow.exceptions import AirflowException
from airflow.models import BaseOperatorLink, DagBag, DagModel, DagRun, TaskInstance
from airflow.operators.dummy import DummyOperator
from airflow.sensors.base import BaseSensorOperator
from airflow.utils.helpers import build_airflow_url_with_query
from airflow.utils.session import provide_session
from airflow.utils.state import State

DELIMITER = '.'

# TO-DO: add exception in the source base class to ensure the delimiter is not used in the source/table names

class TableIngestionSensor(BaseSensorOperator):
    """
    Waits for table ingestion DAGs to complete for a
    specific execution_date

    :param external_source_tables: A map whose keys are the sources 
        to wait for and the values are a list of regex expressions for each source.
        The regex expressions will be used to find matching DAGs. For example:

        external_source_tables = {
            "SourceX" : ["^ABC.*"], # table name starts with ABC followed by any character
            "SourceY" : [".*ABC$"], # table name ends with ABC preceded by any character
            "SourceZ" : [".*"], # wildcard for matching all tables for the given source
            "SourceK" : ["(.*)([^ABC]$)"], # wildcard for matching all tables for the given source AND don't end with ABC
            "SourceW" : ["ABC"], # table name must match exactly ABC
            "SourceM" : ["^ABC.*", "YXZ"] # table name starts with ABC followed by any character OR table name matches YX
        }

    :type external_source_tables: dict
    :param mode: How the sensor operates.
    :type mode: str
    :param allowed_states: Iterable of allowed states, default is ``['success']``
    :type allowed_states: Iterable
    :param failed_states: Iterable of failed or dis-allowed states, default is ``None``
    :type failed_states: Iterable
    :param execution_delta: time difference with the previous execution to
        look at, the default is the same execution_date as the current task or DAG.
        For yesterday, use [positive!] datetime.timedelta(days=1). Either
        execution_delta or execution_date_fn can be passed to
        ExternalTaskSensor, but not both.
    :type execution_delta: Optional[datetime.timedelta]
    """

    template_fields = ['external_source_tables']
    ui_color = '#19647e'

    def __init__(
        self,
        *,
        external_source_tables: dict,
        mode: Optional[str] = 'reschedule',
        allowed_states: Optional[Iterable[str]] = None,
        failed_states: Optional[Iterable[str]] = None,
        execution_delta: Optional[datetime.timedelta] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.allowed_states = list(allowed_states) if allowed_states else [State.SUCCESS]
        self.failed_states = list(failed_states) if failed_states else []

        total_states = self.allowed_states + self.failed_states
        total_states = set(total_states)

        if set(self.failed_states).intersection(set(self.allowed_states)):
            raise AirflowException(
                f"Duplicate values provided as allowed "
                f"`{self.allowed_states}` and failed states `{self.failed_states}`"
            )
       
        self.execution_delta = execution_delta
        self.external_source_tables = external_source_tables

    @provide_session
    def poke(self, context, session=None):
        if self.execution_delta:
            dttm = context['execution_date'] - self.execution_delta
        else:
            dttm = context['execution_date']

        dttm_filter = [dttm]
        serialized_dttm_filter = ','.join(dt.isoformat() for dt in dttm_filter)

        count_allowed = self.get_count(dttm_filter, context, session, self.allowed_states)

        count_failed = -1
        if self.failed_states:
            count_failed = self.get_count(dttm_filter, context, session, self.failed_states)

        if count_failed == len(dttm_filter):
            pass # TO-DO: define behaviour if any of the dependent DAGs have failed. Maybe provide list of essential tables?

        return count_allowed == 1

    def get_count(self, dttm_filter, context, session, states) -> int:
        """
        Get the count of records against dttm filter and states
        :param dttm_filter: date time filter for execution date
        :type dttm_filter: list
        :param session: airflow session object
        :type session: SASession
        :param states: task or dag states
        :type states: list
        :return: count of record against the filters
        """
        TI = TaskInstance
        DR = DagRun
        
        external_dag_ids = self.get_external_dag_ids(context=context, session=session)
        
        expected_count = len(external_dag_ids)
        count = (
            session.query(func.count())
            .filter(
                DR.dag_id.in_(external_dag_ids),
                DR.state.in_(states),
                DR.execution_date.in_(dttm_filter),
            )
            .scalar()
        )

        self.log.info(
            'Current count of completed DAGs is %s. The expected DAG count is %s',
            count,
            expected_count
        )
            
        return count / expected_count

    def get_external_dag_ids(self, context, session) -> list:
        """
        Retrieve a list of external DAG IDs based on the provided source & table combinations
        """
        schedule_interval = context['dag'].schedule_interval

        external_dag_ids = []

        # Query all active dags
        query = session.query(DagModel).filter(DagModel.is_active==True).all()

        if len(query) == 0:
            raise AirflowException(f'No active dags found.')

        schedule_map = {}
        source_dag_map = {}
        for dag in query:
            dag_id = dag.dag_id
            schedule_map[dag_id] = dag.schedule_interval

            if len(dag_id.split(DELIMITER)) != 2:
                continue

            source = dag_id.split(DELIMITER)[0]
            
            if source in source_dag_map:
                source_dag_map[source].append(dag_id)
            else:
                source_dag_map[source] = [dag_id]
        
        if len(source_dag_map) == 0:
            raise AirflowException(f'Unable to determine table ingestion DAGs. Make sure the period delimiter is used correctly.')

        for source, tables in self.external_source_tables.items():
            source_dags = source_dag_map.get(source, None)

            if not source_dags:
                raise AirflowException(f'No active dags found for source {source}.')
            
            dags = []
            for regex in tables:
                try:
                    table_dags = [
                        dag for dag in source_dags if re.match(regex, dag.split(DELIMITER)[1])
                    ]
                except re.error:
                    raise AirflowException(f'The regex expression \'{regex}\' is invalid.')

                if not table_dags:
                    raise AirflowException(f'No active dags found for source {source} using regex: \"{regex}\".')

                if not all([schedule_interval == schedule_map[dag] for dag in table_dags]):
                    raise AirflowException(f'Incompatible schedule intervals with that of the main DAG: {schedule_interval}.')

                dags.extend(table_dags)

            dags = list(set(dags))
            external_dag_ids.extend(dags)

            self.log.info('%s dependent DAGs found for source %s: %s.', len(dags), source, dags)
                        
        return list(set(external_dag_ids))


from airflow.sensors.base import BaseSensorOperator
from airflow.utils.decorators import apply_defaults
from gcp_airflow_foundations.operators.api.hooks.mailgun_hook_test import MailgunHook
from typing import Any

class MailgunValidateSensor(BaseSensorOperator):
    @apply_defaults
    def __init__(self, mailgun_conn: str, job_id: str, **kwargs: Any):
        super().__init__(**kwargs)
        self.job_id = job_id
        self.hook = MailgunHook(mailgun_conn_id=mailgun_conn)

    def poke(self, context):
        res = self.hook.check_job_status(self.job_id)
        return 'download_url' in res.json()


class SqlHelperODS:
    """
    SQL helper class used to formulate the SQL queries for the merge operations of ODS tables.
    
    :param source: Source table name
    :type source: str
    :param target: Target table name
    :type target: str
    :param source_dataset: Source dataset name
    :type source_dataset: str
    :param target_dataset: Target dataset name
    :type target_dataset: str
    :param surrogate_keys: Surrogate keys of the target table
    :type surrogate_keys: list
    :param columns: List of source table columns
    :type columns: list
    :param gcp_conn_id: Airflow GCP connection ID
    :type gcp_conn_id: str
    :param column_mapping: Column mapping
    :type column_mapping: dict
    :param time_partitioning: Time partitioning option for BigQuery target table. One of HOUR, DAY, or MONTH
    :type time_partitioning: str
    :param ods_metadata: User-provided options for ODS metadata column naming
    :type ods_metadata: OdsTableMetadataConfig
    """

    def __init__(
        self,
        source_dataset,
        target_dataset,
        source,
        target,
        surrogate_keys,
        column_mapping,
        columns,
        ods_metadata,
        gcp_conn_id='google_cloud_default'):

        self.source_dataset = source_dataset
        self.target_dataset = target_dataset
        self.source = source
        self.target = target
        self.surrogate_keys = surrogate_keys
        self.column_mapping = column_mapping
        self.ods_metadata = ods_metadata
        self.gcp_conn_id = gcp_conn_id
        self.columns = columns
        self.hash_column_name = ods_metadata.hash_column_name
        self.primary_key_hash_column_name = ods_metadata.primary_key_hash_column_name
        self.ingestion_time_column_name = ods_metadata.ingestion_time_column_name
        self.update_time_column_name = ods_metadata.update_time_column_name

        self.columns_str_source: str = ",".join(["`{}`".format(col) for col in columns])
        self.columns_str_keys: str = ",".join(surrogate_keys)
        self.columns_str_target: str = ",".join(["`{}`".format(self.column_mapping[i]) for i in columns])

    def create_insert_sql(self):
        rows = f"""{self.columns_str_target}, {self.ingestion_time_column_name}, {self.update_time_column_name}, {self.hash_column_name}, {self.primary_key_hash_column_name}"""
        values = f"""{self.columns_str_source}, CURRENT_TIMESTAMP(), CURRENT_TIMESTAMP(), TO_BASE64(MD5(TO_JSON_STRING(S))), TO_BASE64(MD5(ARRAY_TO_STRING([{",".join(["CAST(S.`{}` AS STRING)".format(surrogate_key) for surrogate_key in self.surrogate_keys])}], "")))"""
        return rows, values

    def create_full_sql(self):
        rows, values = self.create_insert_sql()

        return f"""
            SELECT {",".join([f"`{i}` AS `{self.column_mapping[i]}`" for i in self.columns])},
                CURRENT_TIMESTAMP() AS {self.ingestion_time_column_name},
                CURRENT_TIMESTAMP() AS {self.update_time_column_name}, 
                TO_BASE64(MD5(TO_JSON_STRING(S))) AS {self.hash_column_name}, 
                TO_BASE64(MD5(ARRAY_TO_STRING([{",".join(["CAST(S.`{}` AS STRING)".format(surrogate_key) for surrogate_key in self.surrogate_keys])}], ""))) AS {self.primary_key_hash_column_name}
            FROM `{self.source_dataset}.{self.source}` S
        """

    def create_upsert_sql_with_hash(self):
        comma = ","

        rows, values = self.create_insert_sql()

        return f"""
                MERGE `{self.target_dataset}.{self.target}` T
                USING `{self.source_dataset}.{self.source}` S
                ON {' AND '.join(
            [f'T.{self.column_mapping[surrogate_key]}=S.{surrogate_key}' for surrogate_key in self.surrogate_keys])}
                WHEN MATCHED THEN UPDATE
                    SET {(','.join(f'`{self.column_mapping[col]}`=S.`{col}`' for col in self.columns )) + f'{comma}' + f'{self.update_time_column_name}=CURRENT_TIMESTAMP()' + f'{comma}' +  f'{self.hash_column_name}=TO_BASE64(MD5(TO_JSON_STRING(S)))' }
                WHEN NOT MATCHED THEN
                    INSERT ({rows})
                    VALUES ({values})
            """



from abc import ABC, abstractmethod, abstractproperty

from airflow.models.dag import DAG

from gcp_airflow_foundations.base_class.data_source_table_config import DataSourceTablesConfig
from gcp_airflow_foundations.base_class.source_table_config import SourceTableConfig
from gcp_airflow_foundations.base_class.source_config import SourceConfig
from gcp_airflow_foundations.enums.schema_source_type import SchemaSourceType
from gcp_airflow_foundations.common.gcp.load_builder import load_builder
from gcp_airflow_foundations.source_class.schema_source_config import AutoSchemaSourceConfig, GCSSchemaSourceConfig, BQLandingZoneSchemaSourceConfig

import logging

class DagBuilder(ABC):
    """A base DAG builder for creating a list of DAGs for a given source.

    Attributes
    ----------
    sources : list(DagBuilder)
        List of initialized subclasses of this class (dynamically updated during runtime)
    config : DataSourceTablesConfig
        DAG configuration object
    """
    sources = []

    def __init__(self, default_task_args: dict, config: DataSourceTablesConfig):
        self.default_task_args = default_task_args
        self.config = config
        self.validate_extra_options()
        self.set_schema_method_type()
        self.validate_custom_method()
        super().__init__()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if not hasattr(cls, 'source_type'):
            raise AttributeError(f"{cls} source class is missing 'source_type' attribute")
        cls.sources.append(cls)

    def build_dags(self):
        """
        Main DAG building method
        """
        data_source = self.config.source
        logging.info(f"Building DAG for {data_source.name}")

        dags = []
        for table_config in self.config.tables:
            table_default_task_args = self.default_task_args_for_table(
                self.config, table_config
            )
            logging.info(f"table_default_task_args {table_default_task_args}")

            start_date = table_default_task_args["start_date"]

            with DAG(
                dag_id=f"{data_source.name}.{table_config.table_name}",
                description=f"{data_source.name} to BigQuery load for {table_config.table_name}",
                schedule_interval=data_source.ingest_schedule,
                default_args=table_default_task_args,
                render_template_as_native_obj=True
            ) as dag:    

                load_to_bq_landing = self.get_bq_ingestion_task(dag, table_config)

                self.get_datastore_ingestion_task(dag, load_to_bq_landing, data_source, table_config)
            
                dags.append(dag)
                
        extra_dags = self.get_extra_dags()
        if extra_dags is not None and not extra_dags == []:
            dags = dags + self.get_extra_dags()
        return dags

    @abstractmethod
    def get_bq_ingestion_task(self, table_config: SourceTableConfig):
        """Abstract method for the Airflow task that ingests data to the BigQuery staging table"""
        pass

    def get_datastore_ingestion_task(self, dag, preceding_task, data_source: SourceConfig, table_config:SourceTableConfig):
        """Method for the Airflow task group that upserts data to the ODS and HDS tables"""
        load_builder(
            data_source=data_source,
            table_config=table_config,
            schema_config=BQLandingZoneSchemaSourceConfig,
            preceding_task=preceding_task,
            dag=dag
        )

    def get_extra_dags(self):
    # Override if non-table ingestion DAGs are needed for the flow: return them here
        return []

    @abstractmethod
    def set_schema_method_type(self):
        pass

    def get_schema_method_class(self):
        if self.schema_source_type == SchemaSourceType.GCS:
            return GCSSchemaSourceConfig

        elif self.schema_source_type == SchemaSourceType.AUTO:
            return AutoSchemaSourceConfig

        else:
            return

    def validate_custom_method(self):
        if self.schema_source_type== SchemaSourceType.CUSTOM:
            assert self.get_schema_method_class() is not None, 'A custom method for schema parsing must be provided'

    @abstractmethod
    def validate_extra_options(self):
        pass

    def default_task_args_for_table(
        self, config: DataSourceTablesConfig, table_config: SourceTableConfig
    ):
        return {
            **self.default_task_args,
            "start_date": config.table_start_date(table_config)
        }


from pydantic import validator
from pydantic.dataclasses import dataclass

from typing import List, Optional

from pydantic.networks import stricturl

@dataclass
class FTPSourceConfig:
    """
    Attributes:
        file_name_template: a templated file name that applies to all tables. The DAGs will search for one file per table accordingly,
            and no metadata file is required if this field is supplied.
        file_prefix_filtering: option to search for file by prefix (e.g. if the end of file name is a randomly generated string.)
        airflow_date_template: the airflow templated variable to use for loading dated file names: either "ds" or "prev_ds"
        date_format: The date format expected in the file naming convention
        delimiter: delimiter used in the file
        source_format: format of the files
    """
    file_name_template: Optional[str]
    file_prefix_filtering: bool
    date_format: str
    airflow_date_template: str
    delimeter: str
    source_format: str


import os
from re import L
import textwrap
import warnings
from datetime import datetime
from typing import Callable, List, Optional, Sequence, Set, Union
import logging
from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.sensors.base import BaseSensorOperator, poke_mode_only


class GCSObjectPrefixListExistenceSensor(BaseSensorOperator):
    """
    Checks for the existence of a list of files with specified prefixes in Google Cloud Storage.
    :param bucket: The Google Cloud Storage bucket where the object is.
    :type bucket: str
    :param objects: A list of the objects to check in the Google cloud
        storage bucket.
    :type object: list[str]
    :param google_cloud_conn_id: The connection ID to use when
        connecting to Google Cloud Storage.
    :type google_cloud_conn_id: str
    :param delegate_to: The account to impersonate using domain-wide delegation of authority,
        if any. For this to work, the service account making the request must have
        domain-wide delegation enabled.
    :type delegate_to: str
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :type impersonation_chain: Union[str, Sequence[str]]
    """

    template_fields = (
        'bucket',
        'prefixes',
        'impersonation_chain',
    )
    ui_color = '#f0eee4'

    def __init__(
        self,
        *,
        bucket: str,
        prefixes: list,
        google_cloud_conn_id: str = 'google_cloud_default',
        delegate_to: Optional[str] = None,
        impersonation_chain: Optional[Union[str, Sequence[str]]] = None,
        **kwargs,
    ) -> None:

        super().__init__(**kwargs)
        self.bucket = bucket
        self.prefixes = prefixes
        self.google_cloud_conn_id = google_cloud_conn_id
        self.delegate_to = delegate_to
        self.impersonation_chain = impersonation_chain

    def poke(self, context: dict) -> bool:
        for prefix in self.prefixes:
            self.log.info('Sensor checks existence of file with prefix : %s, %s', self.bucket, prefix)
        hook = GCSHook(
            gcp_conn_id=self.google_cloud_conn_id,
            delegate_to=self.delegate_to,
            impersonation_chain=self.impersonation_chain,
        )

        all_files_exist = True
        for prefix in self.prefixes:
            logging.info(object)
            matching = hook.list(bucket_name=self.bucket, prefix=prefix) 
            logging.info(matching)
            if matching == []:
                all_files_exist = False
    
        return all_files_exist


from airflow import DAG
from airflow.exceptions import AirflowException
from airflow.operators.python_operator import PythonOperator
from airflow.utils.task_group import TaskGroup
from airflow.sensors.external_task import ExternalTaskSensor


def dataflow_taskgroup_builder(
    query_schema,
    dataflow_job_params,
    destination_table,
    destination_schema_table,
    table_name,
    system_name,
    create_job_params,
    run_dataflow_job,
    create_table,
    ingest_metadata,
    table_type_casts
) -> TaskGroup:

    """
    Method for returning a Task Group for JDBC->BQ Dataflow DAGS
        Either returns a taskgroup for 
            1) Table ingestion
            2) Schema ingestion
    """

    taskgroup = TaskGroup(group_id="dataflow_taskgroup")

    create_job_parameters = PythonOperator(
        task_id="create_job_parameters",
        op_kwargs={"config_params": dataflow_job_params,
                    "table_name": table_name,
                    "destination_table": destination_table,
                    "destination_schema_table": destination_schema_table,
                    "query_schema": query_schema,
                    "owner": dataflow_job_params["database_owner"]},
        python_callable=create_job_params,
        task_group=taskgroup,
    )

    trigger_dataflow_job = PythonOperator(
        task_id="run_dataflow_job_to_bq",
        op_kwargs={"template_path": dataflow_job_params["template_path"],
                    "system_name": system_name,
                    "table_name": table_name,
                    "query_schema": query_schema},
        python_callable=run_dataflow_job,
        task_group=taskgroup,
        pool=dataflow_job_params["connection_pool"]
    )

    if not query_schema:

        create_table = PythonOperator(
            task_id="create_table_if_needed",
            op_kwargs={"destination_table": destination_table,
            "schema_table": destination_schema_table,
            "source_table": table_name,
            "table_type_casts": table_type_casts},
            python_callable=create_table,
            task_group=taskgroup,
        )

        if ingest_metadata:
            schema_task_sensor = ExternalTaskSensor(
                task_id="check_bq_schema_updated",
                external_dag_id=f"{system_name}_upload_schema",
                task_group=taskgroup,
            )
            schema_task_sensor >> create_job_parameters >> create_table >> trigger_dataflow_job
        else:
            create_job_parameters >> create_table >> trigger_dataflow_job

    else:
        create_job_parameters >> trigger_dataflow_job

    return taskgroup


import csv
import tempfile
import warnings
import time
from typing import Any, Dict, List, Optional, Sequence, Union
from random import shuffle
import pandas as pd
from datetime import datetime
from dateutil.relativedelta import relativedelta

import pyarrow.parquet as pq
import pyarrow

from airflow.exceptions import AirflowException

from gcp_airflow_foundations.operators.facebook.hooks.ads import CustomFacebookAdsReportingHook
from gcp_airflow_foundations.enums.facebook import AccountLookupScope, ApiObject

from airflow.models import BaseOperator, Variable
from airflow.contrib.hooks.bigquery_hook import BigQueryHook

from google.cloud import bigquery


class FacebookAdsReportToBqOperator(BaseOperator):
    """
    Fetches the results from the Facebook Ads API as desired in the params and fields
    Converts to parquet format and loads to directly BigQuery maintaining the native nested
    representation of the data.

    :param api_object: The API Object to query from
    :type api_object: ApiObject
    :param gcp_project: The Google Cloud Platform project ID
    :type gcp_project: str
    :param account_lookup_scope: Whether to query all or only the active accounts managed by the user.
    :type account_lookup_scope: AccountLookupScope
    :param destination_project_dataset_table: BigQuery staging zone table. String in dotted (<project>.)<dataset>.<table> format.
    :type destination_project_dataset_table: str
    :param accounts_bq_table: BigQuery table with the Facebook Account IDs to query data from. String in dotted (<project>.)<dataset>.<table> format.
    :type accounts_bq_table: str
    :param time_range: Time range used in the Graph API query.
    :type time_range: Dict[str, Any] 
    :param gcp_conn_id: Airflow Google Cloud connection ID
    :type gcp_conn_id: str
    :param facebook_conn_id: Airflow Facebook Ads connection ID
    :type facebook_conn_id: str
    :param api_version: The version of Facebook API. Default to None. If it is None,
        it will use the Facebook business SDK default version.
    :type api_version: str
    :param fields: List of fields that is obtained from Facebook. Found in AdsInsights.Field class.
        https://developers.facebook.com/docs/marketing-api/insights/parameters/v6.0
    :type fields: List[str]
    :param parameters: Parameters that determine the query for Facebook
        https://developers.facebook.com/docs/marketing-api/insights/parameters/v6.0
    :type parameters: Dict[str, Any]
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :type impersonation_chain: Union[str, Sequence[str]]
    """

    template_fields = (
        "facebook_conn_id",
        "impersonation_chain",
        "parameters",
    )

    def __init__(
        self,
        *,
        api_object: ApiObject,
        gcp_project: str,
        account_lookup_scope: AccountLookupScope,
        destination_project_dataset_table: str,
        accounts_bq_table: str,
        fields: List[str],
        parameters: Dict[str, Any] = None,
        time_range: Dict[str, Any] = None,
        api_version: Optional[str] = None,
        gcp_conn_id: str = "google_cloud_default",
        facebook_conn_id: str = "facebook_custom",
        impersonation_chain: Optional[Union[str, Sequence[str]]] = None,
        **kwargs,
    ) -> None:
        super(FacebookAdsReportToBqOperator, self).__init__(
            **kwargs
        )

        self.api_object = api_object
        self.gcp_project = gcp_project
        self.account_lookup_scope = account_lookup_scope
        self.destination_project_dataset_table = destination_project_dataset_table
        self.accounts_bq_table = accounts_bq_table
        self.gcp_conn_id = gcp_conn_id
        self.facebook_conn_id = facebook_conn_id
        self.api_version = api_version
        self.fields = fields
        self.parameters = parameters
        self.time_range = time_range
        self.impersonation_chain = impersonation_chain

    def execute(
        self, 
        context: dict
    ):

        dag_id = context['dag'].dag_id
        ds = context['ds']

        interval_start = datetime.strptime(ds, '%Y-%m-%d')
        interval_end = interval_start + relativedelta(day=31)
        
        if not self.time_range:
            self.parameters['time_range'] = {'since':ds, 'until':ds}
        else:
            self.parameters['time_range'] = {'since':self.time_range['since'], 'until':ds}

        self.log.info("Currently loading data for date range: %s", self.parameters['time_range'])

        service = CustomFacebookAdsReportingHook(
            facebook_conn_id=self.facebook_conn_id, api_version=self.api_version
        )

        if self.account_lookup_scope == AccountLookupScope.ALL:
            facebook_acc_ids = service.get_all_accounts()

        elif self.account_lookup_scope == AccountLookupScope.ACTIVE:
            facebook_acc_ids = service.get_active_accounts_from_bq(
                project_id=self.gcp_project, 
                table_id=self.accounts_bq_table
            )

        shuffle(facebook_acc_ids)

        converted_rows = []
        while True:
            for facebook_acc_id in facebook_acc_ids:

                self.log.info("Currently loading data from Account ID: %s", facebook_acc_id)
            
                try:
                    if self.api_object == ApiObject.INSIGHTS:
                        rows = service.bulk_facebook_report_async(facebook_acc_id=facebook_acc_id, params=self.parameters, fields=self.fields)
                        if rows == -1:
                            self.log.info("Rate Limit has reached 75%. Moving on to the next account. Will retry later")
                            continue

                    elif self.api_object == ApiObject.CAMPAIGNS:
                        rows = service.get_campaigns(facebook_acc_id=facebook_acc_id, params=self.parameters)

                    elif self.api_object == ApiObject.ADSETS:
                        rows = service.get_adsets(facebook_acc_id=facebook_acc_id, params=self.parameters)

                    converted_rows.extend(rows)

                    facebook_acc_ids.remove(facebook_acc_id)

                    self.log.info("Extracting data for account %s completed", facebook_acc_id)
                except:
                    self.log.info("Extracting data for account %s failed. Will retry later.", facebook_acc_id)
                
            if len(facebook_acc_ids) == 0:
                break

        self.log.info("Facebook Returned %s data points", len(converted_rows))

        self.transform_data_types(converted_rows)

        df = pd.DataFrame.from_dict(converted_rows)

        writer = pyarrow.BufferOutputStream()
        pq.write_table(
            pyarrow.Table.from_pandas(df),
            writer,
            use_compliant_nested_type=True
        )
        reader = pyarrow.BufferReader(writer.getvalue())

        hook = BigQueryHook(
            gcp_conn_id=self.gcp_conn_id
        )

        client = hook.get_client(project_id=self.gcp_project)

        parquet_options = bigquery.format_options.ParquetOptions()
        parquet_options.enable_list_inference = True

        job_config = bigquery.LoadJobConfig()
        job_config.source_format = bigquery.SourceFormat.PARQUET
        job_config.parquet_options = parquet_options
        job_config.write_disposition='WRITE_TRUNCATE'

        job = client.load_table_from_file(
            reader, f"{self.destination_project_dataset_table}_{ds}", job_config=job_config
        )

    def transform_data_types(
        self, 
        rows
    ):
        """
        Transforms the fields returned by the Facebook API to float or date data types as appropriate.

        :param rows: List of dictionary rows returned by the Facebook API.
        :type rows: List[dict]
        """
        for i in rows:
            i.pop('date_stop')
            i['date_start'] = datetime.strptime(i['date_start'], '%Y-%m-%d').date()
            for j in i:
                if j.endswith('id') or j.endswith('name'):
                    continue
                elif type(i[j]) == str:
                    i[j] = self.get_float(i[j])
                elif type(i[j]) == list:
                    for k in i[j]:
                        for w in k:
                            if (type(k[w]) == str) and (not w.endswith('id')):
                                k[w] = self.get_float(k[w])

    def get_float(
        self, 
        element
    ):
        """
        Attempts to cast a string object into float.

        :param element: Value to be converted to float.
        :type element: str
        """
        try:
            return float(element)
        except ValueError:
            return element


from airflow.exceptions import AirflowException

from dacite import Config
from dataclasses import dataclass, field

from pydantic import validator

from datetime import datetime
from typing import List, Optional

from gcp_airflow_foundations.base_class.ods_metadata_config import OdsTableMetadataConfig
from gcp_airflow_foundations.base_class.hds_metadata_config import HdsTableMetadataConfig
from gcp_airflow_foundations.enums.ingestion_type import IngestionType
from gcp_airflow_foundations.enums.hds_table_type import HdsTableType
from gcp_airflow_foundations.enums.time_partitioning import TimePartitioning

from gcp_airflow_foundations.base_class.ods_table_config import OdsTableConfig
from gcp_airflow_foundations.base_class.hds_table_config import HdsTableConfig
#from gcp_airflow_foundations.base_class.facebook_table_config import FacebookTableConfig

@dataclass
class SourceTableConfig:
    """
    Table configuration data class.

    Attributes:
        table_name : Table name. Used for Dag Id.
        ingestion_type: FULL or INCREMENTAL.
        landing_zone_table_name_override: Optional staging zone table name.
        dest_table_override: Optional target table name. If None, use table_name instead.
        surrogate_keys : Keys used to identify unique records when merging into ODS.
        column_mapping : Mapping used to rename columns.
        ods_config : ODS table configuration. See :class:`gcp_airflow_foundations.base_class.ods_table_config.OdsTableConfig`.
        hds_config : HDS table configuration. See :class:`gcp_airflow_foundations.base_class.hds_table_config.HdsTableConfig`.
        facebook_table_config: Extra options for ingesting data from the Facebook API.
        extra_options: Field for storing additional configuration options.
        version : The Dag version for the table. Can be incremented if logic changes.
        catchup : Passed to a dag [see doc](https://airflow.apache.org/docs/apache-airflow/stable/dag-run.html#catchup).
            Defaults to True. May want to change it to False if Dag version is changed, and we don't want to rerun past dags.
    """

    table_name: str
    ingestion_type: IngestionType # FULL or INCREMENTAL
    landing_zone_table_name_override: Optional[str]
    dest_table_override: Optional[str]
    surrogate_keys: List[str]
    column_mapping: Optional[dict]
    hds_config: Optional[HdsTableConfig]
    #facebook_table_config: Optional[FacebookTableConfig]
    extra_options: dict = field(default_factory=dict)
    ods_config: Optional[OdsTableConfig] = OdsTableConfig(ods_metadata=OdsTableMetadataConfig())
    version: int = 1
    catchup: bool = True


    # Override values
    def __post_init__(self):
        if self.landing_zone_table_name_override is None:
            self.landing_zone_table_name_override = self.table_name
        if self.extra_options is None:
            self.extra_options = {}
            

    @validator("table_name")
    def valid_source_table(cls, v):
        assert v, "Source table name must not be empty"
        assert "." not in v, "Source table name cannot contain the period character"
        return v


from dataclasses import fields
from os import X_OK
from urllib.parse import urlparse
from dacite import from_dict
from dataclasses import dataclass

from airflow.operators.dummy import DummyOperator
from airflow.providers.google.cloud.sensors.gcs import GCSObjectExistenceSensor

from gcp_airflow_foundations.operators.api.sensors.gcs_sensor import GCSObjectListExistenceSensor
from gcp_airflow_foundations.operators.api.sensors.gcs_prefix_sensor import GCSObjectPrefixListExistenceSensor
from gcp_airflow_foundations.source_class.ftp_source import GenericFileIngestionDagBuilder
from gcp_airflow_foundations.common.gcp.load_builder import load_builder


class GCSFileIngestionDagBuilder(GenericFileIngestionDagBuilder):
    """
    Builds DAGs to load files from GCS to a BigQuery Table.

    For GCS->BQ ingestion, either a metadata file is required or the field templated_file_name must be provided. 
    If a metadata file is provided, itt can be a fixed file, or can be a new file supplied daily.
    Airflow context variables are supported for the file naming, e.g.
        TABLE_METADATA_FILE_{{ ds }}.csv
    for a metadata file supplied daily.

    The format of the metadata file should be a csv with one column as follows:
        FILE_NAME_1
        ...
        FILE_NAME_N
    with all files to ingest
    """
    source_type = "GCS"

    def flag_file_sensor(self, table_config, taskgroup):
        if "flag_file_path" in table_config.extra_options.get("ftp_table_config"):
            flag_file_path = table_config.extra_options.get("ftp_table_config")["flag_file_path"]
            bucket = self.config.source.extra_options["gcs_bucket"]
            return GCSObjectExistenceSensor(
                task_id="wait_for_flag_file",
                bucket=bucket,
                object=flag_file_path,
                task_group=taskgroup
            )
        else:
            return None
            #return DummyOperator(
            #    task_id="dummy_flag_file_sensor",
            #    task_group=taskgroup
            #)

    def file_ingestion_task(self, table_config, taskgroup):
        """
        No ingestion is needed - data is already in GCS, so return a dummy operator.
        """
        return None
        #return DummyOperator(
        #    task_id="dummy_file_ingestion_operator",
        #    task_group=taskgroup
        #)

    def file_sensor(self, table_config, taskgroup):
        """
        Returns an Airflow sensor that waits for the list of files specified by the metadata file provided.
        """
        bucket = self.config.source.extra_options["gcs_bucket"]
        files_to_wait_for = "{{ ti.xcom_pull(key='file_list', task_ids='ftp_taskgroup.get_file_list') }}"

        if self.config.source.extra_options["ftp_source_config"]["file_prefix_filtering"]:
            return GCSObjectPrefixListExistenceSensor(
                task_id="wait_for_files_to_ingest",
                bucket=bucket,
                prefixes=files_to_wait_for,
                task_group=taskgroup
            )
        else:
            return GCSObjectListExistenceSensor(
                task_id="wait_for_files_to_ingest",
                bucket=bucket,
                objects=files_to_wait_for,
                task_group=taskgroup
            )
            


#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains SFTP hook."""
import datetime
import stat
import warnings
from typing import Dict, List, Optional, Tuple

import pysftp
import tenacity
from paramiko import SSHException

from airflow.providers.ssh.hooks.ssh import SSHHook


class SFTPHook(SSHHook):
    """
    This hook is inherited from SSH hook. Please refer to SSH hook for the input
    arguments.

    Interact with SFTP.

    :Pitfalls::

        - In contrast with FTPHook describe_directory only returns size, type and
          modify. It doesn't return unix.owner, unix.mode, perm, unix.group and
          unique.
        - retrieve_file and store_file only take a local full path and not a
           buffer.
        - If no mode is passed to create_directory it will be created with 777
          permissions.

    Errors that may occur throughout but should be handled downstream.

    For consistency reasons with SSHHook, the preferred parameter is "ssh_conn_id".
    Please note that it is still possible to use the parameter "ftp_conn_id"
    to initialize the hook, but it will be removed in future Airflow versions.

    :param ssh_conn_id: The :ref:`sftp connection id<howto/connection:sftp>`
    :type ssh_conn_id: str
    :param ftp_conn_id (Outdated): The :ref:`sftp connection id<howto/connection:sftp>`
    :type ftp_conn_id: str
    """

    conn_name_attr = 'ssh_conn_id'
    default_conn_name = 'sftp_default'
    conn_type = 'sftp'
    hook_name = 'SFTP'

    @staticmethod
    def get_ui_field_behaviour() -> Dict:
        return {
            "hidden_fields": ['schema'],
            "relabeling": {
                'login': 'Username',
            },
        }

    def __init__(
        self,
        ssh_conn_id: Optional[str] = 'sftp_default',
        ftp_conn_id: Optional[str] = 'sftp_default',
        *args,
        **kwargs,
    ) -> None:

        if ftp_conn_id:
            warnings.warn(
                'Parameter `ftp_conn_id` is deprecated.' 'Please use `ssh_conn_id` instead.',
                DeprecationWarning,
                stacklevel=2,
            )
            kwargs['ssh_conn_id'] = ftp_conn_id
        self.ssh_conn_id = ssh_conn_id
        super().__init__(*args, **kwargs)

        self.conn = None
        self.private_key_pass = None
        self.ciphers = None

        # Fail for unverified hosts, unless this is explicitly allowed
        self.no_host_key_check = False

        if self.ssh_conn_id is not None:
            conn = self.get_connection(self.ssh_conn_id)
            if conn.extra is not None:
                extra_options = conn.extra_dejson

                # For backward compatibility
                # TODO: remove in Airflow 2.1

                if 'private_key_pass' in extra_options:
                    warnings.warn(
                        'Extra option `private_key_pass` is deprecated.'
                        'Please use `private_key_passphrase` instead.'
                        '`private_key_passphrase` will precede if both options are specified.'
                        'The old option `private_key_pass` will be removed in Airflow 2.1',
                        DeprecationWarning,
                        stacklevel=2,
                    )
                self.private_key_pass = extra_options.get(
                    'private_key_passphrase', extra_options.get('private_key_pass')
                )

                if 'ignore_hostkey_verification' in extra_options:
                    warnings.warn(
                        'Extra option `ignore_hostkey_verification` is deprecated.'
                        'Please use `no_host_key_check` instead.'
                        'This option will be removed in Airflow 2.1',
                        DeprecationWarning,
                        stacklevel=2,
                    )
                    self.no_host_key_check = (
                        str(extra_options['ignore_hostkey_verification']).lower() == 'true'
                    )

                if 'no_host_key_check' in extra_options:
                    self.no_host_key_check = str(extra_options['no_host_key_check']).lower() == 'true'

                if 'ciphers' in extra_options:
                    self.ciphers = extra_options['ciphers']

    @tenacity.retry(
        stop=tenacity.stop_after_delay(10),
        wait=tenacity.wait_exponential(multiplier=1, max=10),
        retry=tenacity.retry_if_exception_type(SSHException),
        reraise=True,
    )
    def get_conn(self) -> pysftp.Connection:
        """Returns an SFTP connection object"""
        if self.conn is None:
            cnopts = pysftp.CnOpts()
            if self.no_host_key_check:
                cnopts.hostkeys = None
            else:
                if self.host_key is not None:
                    cnopts.hostkeys.add(self.remote_host, self.host_key.get_name(), self.host_key)
                else:
                    pass  # will fallback to system host keys if none explicitly specified in conn extra

            cnopts.compression = self.compress
            cnopts.ciphers = self.ciphers
            conn_params = {
                'host': self.remote_host,
                'port': self.port,
                'username': self.username,
                'cnopts': cnopts,
            }
            if self.password and self.password.strip():
                conn_params['password'] = self.password
            if self.pkey:
                conn_params['private_key'] = self.pkey
            elif self.key_file:
                conn_params['private_key'] = self.key_file
            if self.private_key_pass:
                conn_params['private_key_pass'] = self.private_key_pass

            self.conn = pysftp.Connection(**conn_params)
        return self.conn

    def close_conn(self) -> None:
        """Closes the connection"""
        if self.conn is not None:
            self.conn.close()
            self.conn = None

    def describe_directory(self, path: str) -> Dict[str, Dict[str, str]]:
        """
        Returns a dictionary of {filename: {attributes}} for all files
        on the remote system (where the MLSD command is supported).

        :param path: full path to the remote directory
        :type path: str
        """
        conn = self.get_conn()
        flist = conn.listdir_attr(path)
        files = {}
        for f in flist:
            modify = datetime.datetime.fromtimestamp(f.st_mtime).strftime('%Y%m%d%H%M%S')
            files[f.filename] = {
                'size': f.st_size,
                'type': 'dir' if stat.S_ISDIR(f.st_mode) else 'file',
                'modify': modify,
            }
        return files

    def list_directory(self, path: str) -> List[str]:
        """
        Returns a list of files on the remote system.

        :param path: full path to the remote directory to list
        :type path: str
        """
        conn = self.get_conn()
        files = conn.listdir(path)
        return files

    def create_directory(self, path: str, mode: int = 777) -> None:
        """
        Creates a directory on the remote system.

        :param path: full path to the remote directory to create
        :type path: str
        :param mode: int representation of octal mode for directory
        """
        conn = self.get_conn()
        conn.makedirs(path, mode)

    def delete_directory(self, path: str) -> None:
        """
        Deletes a directory on the remote system.

        :param path: full path to the remote directory to delete
        :type path: str
        """
        conn = self.get_conn()
        conn.rmdir(path)

    def retrieve_file(self, remote_full_path: str, local_full_path: str) -> None:
        """
        Transfers the remote file to a local location.
        If local_full_path is a string path, the file will be put
        at that location

        :param remote_full_path: full path to the remote file
        :type remote_full_path: str
        :param local_full_path: full path to the local file
        :type local_full_path: str
        """
        conn = self.get_conn()
        conn.get(remote_full_path, local_full_path)

    def retrieve_directory(self, remote_full_path: str, local_full_path: str) -> None:
        """
        Transfers the remote directory to a local location.
        If local_full_path is a string path, the directory will be put
        at that location

        :param remote_full_path: full path to the remote directory
        :type remote_full_path: str
        :param local_full_path: full path to the local directory
        :type local_full_path: str
        """
        conn = self.get_conn()
        conn.get_r(remote_full_path, local_full_path)
        

    def store_file(self, remote_full_path: str, local_full_path: str) -> None:
        """
        Transfers a local file to the remote location.
        If local_full_path_or_buffer is a string path, the file will be read
        from that location

        :param remote_full_path: full path to the remote file
        :type remote_full_path: str
        :param local_full_path: full path to the local file
        :type local_full_path: str
        """
        conn = self.get_conn()
        conn.put(local_full_path, remote_full_path)

    def delete_file(self, path: str) -> None:
        """
        Removes a file on the FTP Server

        :param path: full path to the remote file
        :type path: str
        """
        conn = self.get_conn()
        conn.remove(path)

    def get_mod_time(self, path: str) -> str:
        """
        Returns modification time.

        :param path: full path to the remote file
        :type path: str
        """
        conn = self.get_conn()
        ftp_mdtm = conn.stat(path).st_mtime
        return datetime.datetime.fromtimestamp(ftp_mdtm).strftime('%Y%m%d%H%M%S')

    def path_exists(self, path: str) -> bool:
        """
        Returns True if a remote entity exists

        :param path: full path to the remote file or directory
        :type path: str
        """
        conn = self.get_conn()
        return conn.exists(path)

    @staticmethod
    def _is_path_match(path: str, prefix: Optional[str] = None, delimiter: Optional[str] = None) -> bool:
        """
        Return True if given path starts with prefix (if set) and ends with delimiter (if set).

        :param path: path to be checked
        :type path: str
        :param prefix: if set path will be checked is starting with prefix
        :type prefix: str
        :param delimiter: if set path will be checked is ending with suffix
        :type delimiter: str
        :return: bool
        """
        if prefix is not None and not path.startswith(prefix):
            return False
        if delimiter is not None and not path.endswith(delimiter):
            return False
        return True

    def get_tree_map(
        self, path: str, prefix: Optional[str] = None, delimiter: Optional[str] = None
    ) -> Tuple[List[str], List[str], List[str]]:
        """
        Return tuple with recursive lists of files, directories and unknown paths from given path.
        It is possible to filter results by giving prefix and/or delimiter parameters.

        :param path: path from which tree will be built
        :type path: str
        :param prefix: if set paths will be added if start with prefix
        :type prefix: str
        :param delimiter: if set paths will be added if end with delimiter
        :type delimiter: str
        :return: tuple with list of files, dirs and unknown items
        :rtype: Tuple[List[str], List[str], List[str]]
        """
        conn = self.get_conn()
        files, dirs, unknowns = [], [], []  # type: List[str], List[str], List[str]

        def append_matching_path_callback(list_):
            return lambda item: list_.append(item) if self._is_path_match(item, prefix, delimiter) else None

        conn.walktree(
            remotepath=path,
            fcallback=append_matching_path_callback(files),
            dcallback=append_matching_path_callback(dirs),
            ucallback=append_matching_path_callback(unknowns),
            recurse=True,
        )

        return files, dirs, unknowns

    def test_connection(self) -> Tuple[bool, str]:
        """Test the SFTP connection by checking if remote entity '/some/path' exists"""
        try:
            conn = self.get_conn()
            conn.pwd
            return True, "Connection successfully tested"
        except Exception as e:
            return False, str(e)


import json
import logging
import uuid
from datetime import datetime

from airflow.contrib.hooks.bigquery_hook import BigQueryHook
from airflow.contrib.hooks.gcs_hook import (
    GoogleCloudStorageHook
)

from google.cloud import bigquery

class SchemaMigrationAudit:
    """
    Inserts the the audit log rows of the schema migration operations that are executed by the MigrateSchema class. The table is created if it does not exist.
    
    :param project_id: GCP project ID  
    :type project_id: str
    :param table_id: Target table name
    :type table_id: str
    :param dataset_id: Target dataset name
    :type dataset_id: str
    :param delegate_to: The account to impersonate using domain-wide delegation of authority, if any
    :type delegate_to: str
    :param gcp_conn_id: Airflow GCP connection ID
    :type gcp_conn_id: str
    """

    def __init__(
        self,
        project_id,
        dataset_id,
        table_id='schema_migration_audit_table',
        gcp_conn_id='google_cloud_default',
        delegate_to=None
    ):

        self.project_id = project_id
        self.dataset_id = dataset_id
        self.table_id = table_id
        self.gcp_conn_id = gcp_conn_id
        self.delegate_to = delegate_to

        self.migration_id = uuid.uuid4().hex
        self.migration_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        logging.info("Creating audit table in BQ.")
        self.create_audit_table()


    def insert_change_log_rows(
        
        self, 
        change_log: list
    ):
        """
        :param change_log: The schema migration operations executed by the MigrateSchema class
        :type change_log: list
        """
        
        for i in change_log:
            i['schema_updated_at'] = self.migration_time
            i['migration_id'] = self.migration_id
            
        client = bigquery.Client(project=self.project_id)

        table_ref = client.dataset(self.dataset_id).table(self.table_id)
        table = client.get_table(table_ref)

        results = client.insert_rows(table, change_log)
               

    def create_audit_table(self):
            bq_hook = BigQueryHook(gcp_conn_id=self.gcp_conn_id, delegate_to=self.delegate_to)
            conn = bq_hook.get_conn()
            cursor = conn.cursor()

            cursor.create_empty_table(
                project_id=self.project_id,
                dataset_id=self.dataset_id,
                table_id=self.table_id,
                schema_fields=self.__audit_table_schema_fields(),
                exists_ok=True
            )


    def __audit_table_schema_fields(self):
        return [
            {
                "name":"table_id",
                "type":"STRING"
            },
            {
                "name":"dataset_id",
                "type":"STRING"
            },
            {
                "name":"schema_updated_at",
                "type":"TIMESTAMP"
            },
            {
                "name":"migration_id",
                "type":"STRING"
            },
            {
                "name":"column_name",
                "type":"STRING"
            },
            {
                "name":"type_of_change",
                "type":"STRING"
            }
        ]



import os
import sys

import setuptools
from setuptools import setup, find_packages

from pathlib import Path

here = os.path.abspath(os.path.dirname(__file__))
about = {}

with open(os.path.join(here, "requirements.txt"), "r") as f:
	requirements = f.read().strip().split('\n')

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

packages = [
    package
    for package in setuptools.PEP420PackageFinder.find()
    if package.startswith("gcp_airflow_foundations")
]

version = {}
with open(os.path.join(here, "gcp_airflow_foundations/version.py")) as fp:
    exec(fp.read(), version)
version = version["__version__"]

def main():
    metadata = dict(
        name="gcp-airflow-foundations-dev",
        version=version,
        description="Opinionated framework based on Airflow 2.0 for building pipelines to ingest data into a BigQuery data warehouse",
        long_description=long_description,
        long_description_content_type='text/markdown',
        url="https://github.com/badal-io/gcp-airflow-foundations",
        author="Badal.io",
        author_email="info@badal.io",
        license="Apache 2.0",
        packages=packages,
        install_requires=requirements,
        classifiers=[
            'Development Status :: 4 - Beta',      
            'Intended Audience :: Developers',      
            'Topic :: Software Development :: Libraries',
            'License :: OSI Approved :: Apache Software License',   
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
        ]
    )

    setup(**metadata)


if __name__ == "__main__":
    main()



from pydantic import validator
from pydantic.dataclasses import dataclass

from typing import List, Optional

@dataclass
class SFTPTableConfig:
    """
    Attributes:
        sftp_connection_name: name of the connection in Airflow
        date_column: optional date column name for scanning/replacing in external partitions for .PARQUET uploads
    """
    sftp_connection_name: str
    sftp_private_key_secret_name: Optional[str]
    date_column: Optional[str]
    


import time
from typing import Any, Dict, List, Optional
from enum import Enum
from datetime import datetime
import requests
import json

from airflow.exceptions import AirflowException
from airflow.providers.facebook.ads.hooks.ads import FacebookAdsReportingHook

from facebook_business.api import FacebookAdsApi
from facebook_business.adobjects.adreportrun import AdReportRun
from facebook_business.adobjects.adaccount import AdAccount
from facebook_business.adobjects.adsinsights import AdsInsights

from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook
from google.cloud import bigquery


class JobStatus(Enum):
    """Available options for facebook async task status"""

    COMPLETED = 'Job Completed'
    STARTED = 'Job Started'
    RUNNING = 'Job Running'
    FAILED = 'Job Failed'
    SKIPPED = 'Job Skipped'


class CustomFacebookAdsReportingHook(FacebookAdsReportingHook):
    """
    Custom Hook for the Facebook Ads API. It extends the default FacebookAdsReportingHook.

    :param facebook_conn_id: Airflow Facebook Ads connection ID
    :type facebook_conn_id: str
    :param api_version: The version of Facebook API. Default to None. If it is None,
        it will use the Facebook business SDK default version.
    :type api_version: Optional[str]
    """

    conn_name_attr = 'facebook_conn_id'
    default_conn_name = 'facebook_custom'

    def __init__(
        self,
        facebook_conn_id: str = default_conn_name,
        api_version: Optional[str] = None,
        **kwargs
    ) -> None:
        super(CustomFacebookAdsReportingHook, self).__init__(
            facebook_conn_id=facebook_conn_id,
            api_version=api_version,
            **kwargs
        )

        self.facebook_conn_id = facebook_conn_id
        self.api_version = api_version
        self.client_required_fields = ["app_id", "app_secret", "access_token"]
        self.config = self.facebook_ads_config      

    def _get_service(
        self, 
        facebook_acc_id
    ) -> FacebookAdsApi:
        """Returns Facebook Ads Client using a service account"""

        return FacebookAdsApi.init(
            app_id=self.config["app_id"],
            app_secret=self.config["app_secret"],
            access_token=self.config["access_token"],
            account_id=facebook_acc_id,
            api_version=self.api_version,
        )

    def bulk_facebook_report_async(
        self,
        facebook_acc_id: str,
        params: Dict[str, Any],
        fields: List[str],
        sleep_time: int = 5,
    ) -> List[AdsInsights]:
        """
        Pulls data from the Facebook Ads API using async calls.
        :param facebook_acc_id: The Facebook account ID to pull data from.
        :type facebook_acc_id: str
        :param fields: List of fields that is obtained from Facebook. Found in AdsInsights.Field class.
            https://developers.facebook.com/docs/marketing-api/insights/parameters/v6.0
        :type fields: List[str]
        :param params: Parameters that determine the query for Facebook
            https://developers.facebook.com/docs/marketing-api/insights/parameters/v6.0
        :type fields: Dict[str, Any]
        :param sleep_time: Time to sleep when async call is happening
        :type sleep_time: int
        :return: Facebook Ads API response, converted to rows.
        :rtype: List[dict]
        """

        api = self._get_service(facebook_acc_id=facebook_acc_id)
        ad_account = AdAccount(api.get_default_account_id(), api=api)
        _async = ad_account.get_insights(params=params, fields=fields, is_async=True)
        while True:
            request = _async.api_get()
            async_status = request[AdReportRun.Field.async_status]
            percent = request[AdReportRun.Field.async_percent_completion]
            self.log.info("%s %s completed, async_status: %s", percent, "%", async_status)
            if async_status == JobStatus.COMPLETED.value:
                self.log.info("Job run completed")
                break
            if async_status in [JobStatus.SKIPPED.value, JobStatus.FAILED.value]:
                message = f"{async_status}. Please retry."
                raise AirflowException(message)
            time.sleep(sleep_time)
        report_run_id = _async.api_get()["report_run_id"]
        report_object = AdReportRun(report_run_id, api=api)
        insights = report_object.get_insights()

        max_current_usage = self.usage_throttle(insights)

        if max_current_usage >= 75:
            return -1

        self.log.info("Extracting data from returned Facebook Ads Iterators")

        rows = []
        while True:
            max_current_usage = self.usage_throttle(insights)
            if max_current_usage >= 75:
                self.log.info('75% Rate Limit Reached. Cooling Time 5 Minutes.')
                time.sleep(300)
            try:
                rows.append( next(insights) )
            except StopIteration:
                break

        return [dict(row) for row in rows]

    def bulk_facebook_report(
        self,
        facebook_acc_id: str,
        params: Dict[str, Any],
        fields: List[str],
        sleep_time: int = 5,
    ) -> List[AdsInsights]:
        """
        Pulls data from the Facebook Ads API using sync calls.
        :param facebook_acc_id: The Facebook account ID to pull data from.
        :type facebook_acc_id: str
        :param fields: List of fields that is obtained from Facebook. Found in AdsInsights.Field class.
            https://developers.facebook.com/docs/marketing-api/insights/parameters/v6.0
        :type fields: List[str]
        :param params: Parameters that determine the query for Facebook
            https://developers.facebook.com/docs/marketing-api/insights/parameters/v6.0
        :type fields: Dict[str, Any]
        :return: Facebook Ads API response, converted to rows.
        :rtype: List[dict]
        """

        api = self._get_service(facebook_acc_id=facebook_acc_id)
        ad_account = AdAccount(api.get_default_account_id(), api=api)
        insights = ad_account.get_insights(params=params, fields=fields, is_async=False)
        rows = list(insights)

        self.usage_throttle(insights)

        return [dict(row) for row in rows]

    def get_campaigns(
        self,
        facebook_acc_id: str,
        params: Dict[str, Any]
    ) -> List[dict]:
        """
        Pulls campaign data from the Facebook Ads API using sync calls.
        :param facebook_acc_id: The Facebook account ID to pull data from.
        :type facebook_acc_id: str
        :param params: Parameters that determine the query for Facebook
            https://developers.facebook.com/docs/marketing-api/insights/parameters/v6.0
        :type params: Dict[str, Any]
        :return: Facebook Ads API response, converted to rows.
        :rtype: List[dict]
        """

        api = self._get_service(facebook_acc_id=facebook_acc_id)
        ad_account = AdAccount(api.get_default_account_id(), api=api)

        campaigns = ad_account.get_campaigns(
            params={'limit':'20000','time_range':params['time_range']},
            fields=[
            'account_id',
            #'name', TO-DO: troubleshoot why pyarrow fails to convert the `name` column
            'daily_budget',
            'effective_status',
            'lifetime_budget',
            'start_time',
            'stop_time'
            ]
        )

        rows = []
        for row in campaigns:
            converted_row = row._data
            if 'name' in converted_row:
                converted_row['name'] = str(converted_row['name'])
            if 'start_time' in converted_row:
                converted_row['start_time'] = datetime.strptime(converted_row['start_time'],'%Y-%m-%dT%H:%M:%S%z')
            if 'stop_time' in row:
                converted_row['stop_time'] = datetime.strptime(converted_row['stop_time'],'%Y-%m-%dT%H:%M:%S%z')
            rows.append(converted_row)

        return rows

    def get_adsets(
        self,
        facebook_acc_id: str,
        params: Dict[str, Any]
    ) -> List[dict]:
        """
        Pulls adset data from the Facebook Ads API using sync calls.
        :param facebook_acc_id: The Facebook account ID to pull data from.
        :type facebook_acc_id: str
        :param params: Parameters that determine the query for Facebook
            https://developers.facebook.com/docs/marketing-api/insights/parameters/v6.0
        :type fields: Dict[str, Any]
        :return: Facebook Ads API response, converted to rows.
        :rtype: List[dict]
        """

        api = self._get_service(facebook_acc_id=facebook_acc_id)
        ad_account = AdAccount(api.get_default_account_id(), api=api)

        adsets = ad_account.get_ad_sets(
            params={'limit':'20000','time_range':params['time_range']},
            fields=[
                'account_id',
                'name',
                'daily_budget',
                'effective_status',
                'lifetime_budget',
                'created_time',
                'end_time'
            ]
        )

        rows = []
        for row in adsets:
            converted_row = row._data
            if 'name' in converted_row:
                converted_row['name'] = str(converted_row['name'])
            if 'created_time' in converted_row:
                converted_row['created_time'] = datetime.strptime(converted_row['created_time'],'%Y-%m-%dT%H:%M:%S%z')
            if 'end_time' in row:
                converted_row['end_time'] = datetime.strptime(converted_row['end_time'],'%Y-%m-%dT%H:%M:%S%z')
            rows.append(converted_row)

        return rows

    def get_active_accounts_from_bq(
        self, 
        project_id, 
        table_id
     ) -> List[str]:
        """
        Pulls a list of Facebook account IDs from a BigQuery table.
        :param project_id: The Google Cloud Platform project ID.
        :type project_id: str
        :param table_id: Name of BigQuery table that contains the Facebook account IDS.
        :type table_id: str
        :return: A list with the Facebook account IDs.
        :rtype: List[str]
        """

        sql = f"SELECT account_id FROM `{table_id}`"

        query_config = bigquery.QueryJobConfig(use_legacy_sql=False)

        client = bigquery.Client(project=project_id)

        df = client.query(sql, job_config=query_config).to_dataframe()

        return [f"act_{i}" for i in df.account_id]

    def get_all_accounts(
        self
    ) -> List[str]:
        """
        Pulls a list of Facebook account IDs from the Facebook API.
        :return: A list with the Facebook account IDs.
        :rtype: List[str]
        """

        self.log.info("Extracting all accounts")

        user_id = self.config["user_id"]

        URL = f"https://graph.facebook.com/v12.0/{user_id}/adaccounts"
        params = {
            "access_token":self.config["access_token"],
            "limit":10000
        }

        accounts = requests.get(URL, params=params).json()['data']

        return [i['id'] for i in accounts]

    def usage_throttle(
        self, 
        insights
    ) -> int:
        """
        Queries the 'x-business-use-case-usage' header of the Cursor object returned by the Facebook API.
        """
        
        usage_header = json.loads(insights._headers['x-business-use-case-usage'])
        values = list(usage_header.values())[0][0]
        return max(values['call_count'], values['total_cputime'], values['total_time'])